////////////////////////////////////////////////////////////////////////////////
// Thu Jan 11 06:26:36 PST 2018
//
// harmonic.cpp: Quantum harmonic oscillator
//
// compile with: g++  -o harmonic harmonic.cpp -llapack
//
// use: ./harmonic N h
//
// This program solves the Schroedinger equation for a harmonic oscillator
// in one dimension. The quadratic potential is V(x) = 0.5 * x * x
// The equation is discretized using a centered finite difference for
// the Laplacian operator
//
// N: number of grid points in the positive x direction (excluding x=0)
//    The total number of grid points used is 2*N+1
// h: grid spacing
//    The range of the grid points is [-N*h, ..., N*h]
//
// The lowest 10 eigenvalues are printed on standard output. The corresponding
// eigenvectors are written on the file "harmonic.dat" in a format suitable 
// for plotting using the Gnuplot program.
//
////////////////////////////////////////////////////////////////////////////////

#include<iostream>
#include<iomanip>
#include<cstdlib>
#include<fstream>
using namespace std;

// Use the BLAS library for matrix diagonalization
extern "C"
void dsyev_(char* jobz, char* uplo, int* n, double* a, int* lda,
            double *w, double *wrk, int* lwrk, int* info);

int main(int argc, char** argv)
{
  if ( argc == 1 )
  {
    cout << "Use: harmonic N h" << endl;
    return 1;
  }
  // input: N = number of grid points in positive x direction, including 0
  const int N = atoi(argv[1]);
  int n = 2*N+1;
  // input: h = grid spacing
  const double h = atof(argv[2]);

  // allocate array for the hamiltonian matrix a
  double *a = new double[n*n];

  // allocate array for the eigenvalue array e
  double *e = new double[n];

  // allocate work space for the DSYEV BLAS library function
  int lwrk = 3*n-1;
  double *wrk = new double[lwrk];

  // initialize the matrix a to zero
  for ( int i = 0; i < n*n; i++ )
    a[i] = 0.0;

  // diagonal elements of the matrix a
  for ( int i = 0; i < n; i++ )
    a[i+n*i] = 1.0/(h*h) + 0.5*((i-N)*h)*((i-N)*h);

  // off-diagonal elements (only the lower triangle is needed)
  for ( int i = 0; i < n-1; i++ )
    a[i+1+n*i] = -0.5/(h*h);
  
  // matrix a is provided in the lower triangle: uplo = 'L'
  char uplo = 'L';
  // compute eigenvalues and eigenvectors: jobz = 'V'
  char jobz = 'V';
  int info;

  // compute eigenvalues of a
  dsyev_(&jobz,&uplo,&n,a,&n,e,wrk,&lwrk,&info);

  // print at most 10 eigenvalues
  cout << "Eigenvalues" << endl;
  for ( int i = 0; (i < 10) && (i < n); i++ )
    cout << setw(12) << setprecision(6) << e[i] << endl;

  // write at most 10 lowest eigenvectors on file harmonic.dat
  ofstream outfile("harmonic.dat");
  for ( int i = 0; (i < 10) && (i < n); i++ )
  {
    outfile << "# eigenvector i=" << i << " e = " << e[i] << endl;
    for ( int j = 0; j < n; j++ )
       outfile << (j-N)*h << " " << a[j+n*i] << endl;
    outfile << endl;
    outfile << endl;
  }
  return 0;
}
